#!/usr/bin/perl -w

# Program entropy.pl

# (c) Gary C. Kessler, 2016 [gck@garykessler.net]

# Program to compute the entropy of a given file

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

# Initialize global variables

$version = "1.5";
$build_date = "30 March 2017";

# We need the log (base 2) of a value but Perl doesn't do log (base 2).
# Instead, we get natural logs (i.e., base e), aka ln
# To calculate log (base 2) X, we computer ln (X) / ln (2)

$ln2 = log (2.0);

# $source is the name of the file to read.

$source = "";
$verbose = "FALSE";
$bar = "FALSE";

#
# Begin program activity!!
#

print "\nFile Entropy Calculator v$version - Gary C. Kessler ($build_date)\n\n";

# Parse command for file name and request missing information.

if (parsecommandline ())
  { exit 1; };

print "Source file = $source.";

# Open input ($source) file and stop program if error

open (INFILE, "<:raw", $source)
  or die "Can't open input file \"$source\": $!";
binmode INFILE;

# Start by reading the entire file, count the number of bytes, and add to
# the frequency table... after initializing these variables!

for ($i=0; $i<256; $i++)
  { $freq [$i] = 0; }
$filesize = 0;

while (not (eof INFILE))
  {
  read (INFILE,$byte,1);
  $temp = unpack ("C", $byte);
  $freq [$temp]++;
  $filesize++;
  }

close INFILE;
printf (" File size = %s bytes.\n\n", commify ($filesize));

if ($bar eq "TRUE")
  { histogram (); }

# Convert the frequency table from a raw count to actual frequency by dividing
# by byte count, then calculate sigma

# Entropy formula (S) = - sigma {i=0,255} freq[i] * log2 (freq[i])

$cum_freq = 0;
$sigma = 0;

for ($i=0; $i<256; $i++)
  {
  $freq [$i] = $freq [$i] / $filesize;
  if ($freq [$i] > 0)
    { $sigma += ($freq [$i] * (log ($freq [$i]) / $ln2)); }
  $cum_freq += $freq [$i] * 100.0;
  }

$sigma = -$sigma;

if ($verbose eq "TRUE")
  {
  print "                                    BYTE VALUE FREQUENCY TABLE\n";
  for ($i=0; $i<256; $i++)
    {
    printf "Value x%02X = %8.5f%%   ", $i, $freq [$i]*100.0;
    if ( ($i+1) % 4 == 0)
      { print "\n"; }
    }
  printf "Frequency sigma = %8.3f%%\n\n", $cum_freq;
  }

print "Entropy = $sigma\n\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# commify
# ********************************************************

sub commify

# Routine from PerlFAQ5 -- add commas to large integers
# http://perldoc.perl.org/perlfaq5.html
# How-can-I-output-my-numbers-with-commas-added?

# Sample usage: printf (" File size = %s bytes\n\n", commify ($count));

{
local $_  = shift;
1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
return $_;
}

# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: entropy [-i input_file] [-b] [-v]
               entropy [-h]

 where: -i is the input file name
        -b displays a histogram (bar graph)
        -h prints this help file
        -v displays verbose output

If no -i switch is provided, the program will ask for the
input file name.

EOT
return;
}

# ********************************************************
# histogram
# ********************************************************

# Display a frequency histogram

sub histogram
{
my ($i, $j, $n, $max, $scale);
my $width = 60;

# Start by finding the largest value in the $freq[ ] array, and set the scale

$max = 0;
for ($i=0; $i<256; $i++)
  {
  if ($freq [$i] > $max)
    { $max = $freq [$i]; }
  }

$scale = $max / $width;

# Print the histogram

print "                                    HISTOGRAM\n";
printf "One X = ~%s occurences\n\n", commify (int($scale+0.5));

for ($i=0; $i<256; $i++)
  {
  printf "x%02X", $i;
  if ($i >= 32 && $i <= 126)
      { printf " %c", $i; }
    else
      { print "  "; }
  print " | ";

  $n = int (($freq [$i] / $scale) + 0.5);
  if ($n > 0)
    {
    for ($j=1; $j<=$n; $j++)
      { print "X"; }
    }
  print "\n";
  }

print "\n";

return;
}


# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)

  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-b" && do
         {
         $bar = "TRUE";
         last PARMS;
         };
      $ARGV[$i] eq "-v" && do
         {
         $verbose = "TRUE";
         last PARMS;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: entropy [-i input_file] [-b] [-v]\n";
         print "       entropy [-h]\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file names, if not supplied

if ($source eq "")

  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

return $state;
}


